# précondition pour que la pc k couvre la mc (si t=1) ou gc (si t=2) l
# t==1 : si xs apparait dans l, alors xr doit apparaître ds k ; x=1,r,s
# t==2 : si xs apparait dans l, alors x1 doit apparaître ds k ; x=1,r,s
def precouvre(k,l,t):
    if t==1:
        return (l[2]==0 or k[1]>0) and (l[4]==0 or k[3]>0) and (l[5]==0 or k[4]>0)
    else:
        return (l[2]==0 or k[0]>0) and (l[4]==0 or k[1]>0) and (l[5]==0 or k[2]>0)

# la pc k suffit-elle à "couvrir" tous les s d'au moins une séquence associée à la mc (si t=1) ou gc (si t=1) l ?
# si t==1 : xsy ds la seq de la mc implique xry ou yrx ds celle de la pc
# si t==2 : xsy ds la seq de la gc implique x1y ou y1x ds celle de la pc
# très long car calcule d'abord toutes les séquences, et il peut y en avoir bcp
def couvre(k,l,t):
    pseq=sequence(k)
    pseq=pseq+pseq[0:2] # pour éviter pb circularité
    voisins=[('1','1'),('1','r'),('1','s'),('r','r'),('r','s'),('s','s')]
    if t==1:
        pxty=[(x+'r'+y in pseq or  y+'r'+x in pseq) for (x,y) in voisins]
    else:
        pxty=[(x+'1'+y in pseq or  y+'1'+x in pseq) for (x,y) in voisins]    
    for tseq in sequences(l):
        tseq=tseq+tseq[0:2] # pour éviter pb circularité
        txsy=[(x+'s'+y in tseq or  y+'s'+x in tseq) for (x,y) in voisins]
        if min(vector(pxty)-vector(txsy))>=0:
            return true
    return false

# résolution des couples pc/mc dans I
# t=0 : AA ; t=1 : CIF
# sortie (RS,E,D1) : solutions, erreurs, dimension1
def resultant(I,t):
    (RS,E,D1)=([],[],[])
    for (i,j) in I:
        try:
            R=ZZ[r](eqpc[i].resultant(eqmc[j],s))
            S=ZZ[s](eqpc[i].resultant(eqmc[j],r))
            # si R ou S nul : systeme de dim >0
            if R==0 or S==0:
                D1.append((i,j))
                print('(%2d, %3d) : dimension 1 ou plus'%(i,j))
            else:
                RR=[]
                SS=[]
                # racines des facteurs pas forcément distinctes : redondance
                for R in [f[0] for f in factor(R) if not inutile(f[0])]:
                    if t==0:
                        RR+=R.roots(AA,multiplicities=false)
                    else:
                        RR+=[x.real() for x in R.roots(CIF,multiplicities=false) if x.imag().contains_zero()]
                for S in [f[0] for f in factor(S) if not inutile(f[0])]:
                    if t==0:
                        SS+=S.roots(AA,multiplicities=false)
                    else:
                        SS+=[x.real() for x in S.roots(CIF,multiplicities=false) if x.imag().contains_zero()]
                z=flatten([[((i,j),(rr,ss)) for rr in RR] for ss in SS],max_level=1)
                RS+=z
                print('(%2d, %3d) : %d couples ajoutés'%(i,j,len(z)))
        except:
            print('(%2d, %3d) : %s'%(i,j,sys.exc_info()))
            E.append((i,j))
    return (RS,E,D1)

def passe2(RS):
    RS2=[]
    for ((i,j),(rr,ss)) in RS:
        if verif_couronne(pc[i],0,rr,ss):
            if verif_couronne(mc[j],1,rr,ss):
                RS2.append(((i,j),(rr,ss)))
    return RS2

def passe3(RS):
    RS2=[]
    c1=0
    c2=0
    for ((i,j),(rr,ss)) in RS:
        print("N°%d : (%2d, %3d)..."%(c1,i,j))
        c1+=1
        p=couronnes_compatibles(rr,ss,0)
        if len(p)==1: # 2pc ou plus : traité ailleurs
            m=couronnes_compatibles(rr,ss,1)
            m=[k for k in m if precouvre(p[0],k,1) and couvre(p[0],k,1)] # un peu utile
            m=[k for k in m if not trou_vide(k,1,rr,ss)] # vire couronne avec interstice remplissable non rempli : le plus utile
            if len(m)>0:
                g=couronnes_compatibles(rr,ss,2)
                g=[k for k in g if precouvre(p[0],k,2)] # sert à peu près à rien
                g=[k for k in g if not trou_vide(k,2,rr,ss)] # vire couronne avec interstice remplissable non rempli : le plus utile
                if len(g)>0:
                    # si plusieurs mc, indexe par celle d'indice min (pas dans I1) pour éviter doublon
                    j=min([mc.index(j) for j in m if j in mc and (i,mc.index(j)) not in I1])
                    if (i,j) not in [k[0] for k in RS2]:
                        RS2.append(((i,j),(rr,ss),(p,m,g)))
                        c2+=1
                        print('%d retenus'%c2)
    return RS2

# vérif exacte des couples ds RS
def passe4(RS):
    E=[]
    for c in RS:
        try:
            verif_equation(c)
        except:
            print('(%d,%d) : ERREUR'%(c[0][0],c[0][1]))
            E.append(c)
    return E
    
#####################
#####################
#####################

# toutes les pc
pc=petites_couronnes()
# toutes les mc
# mcs=[moyennes_couronnes(k) for k in pc[0:10]] # 5min 22s
mcs=load('mcs')

# z=[equation(i,1) for i in mcs[1]] : 2h 21min 47s
# save(z,'eqmcs') # 36,2 Mo

# tous les couples pc/mc
pcmc=[mcs[pc.index(angles(sequence(k).replace('1','r')))] for k in pc] # 16805 couples
# la pc doit contenir un r et la mc un s (sinon biphasé ou autre mc en contact avec la pc)
pcmc=[[l for l in pcmc[i] if pc[i][1]+pc[i][3]+pc[i][4]>0 and l[2]+l[4]+l[5]>0] for i in range(55)] # 12168 couples
# les pc 10, 13, 17 et 37 impose une 2eme pc (autre que ssssss) : cas 2pc
# on les vire -> 10028 couples
#pcmc[10]=[]
#pcmc[13]=[]
#pcmc[17]=[]
#pcmc[37]=[]
pcmc=[[l for l in pcmc[i] if precouvre(pc[i],l,1)] for i in range(55)] # 2889 couples
pcmc=[[l for l in pcmc[i] if couvre(pc[i],l,1)] for i in range(55)] # 1min 19s, 803 couples

mc=list(set(flatten(pcmc,max_level=1))) # 192 couronnes
eqpc=[equation(i,0) for i in pc] # 6.5s
eqmc=[equation(i,1) for i in mc] # 1min 51s
K.inject_variables()
I=flatten([[(i,mc.index(j)) for j in pcmc[i]] for i in range(len(pc))],max_level=1)

#####################
# couples (r,s) via résultant

(RS,E,I1)=resultant(I,0) # 469808 couples en 3min 51s, 5 erreurs, 12 cas dim>0
(RSE,E2,I1E)=resultant(E,1) # 43448 couples en 1min 9s

#####################
# passe 1 : domaine
# comparaison : "vrai" si on ne peut pas décider (ce n'est pas le comportement par défaut)
RS1=[(k,(rr,ss)) for (k,(rr,ss)) in RS if 0<RIF(ss).endpoints()[1] and RIF(ss).endpoints()[0]<RIF(rr).endpoints()[1] and RIF(rr).endpoints()[0]<1] # 57017 en 20s
RS1E=[(k,(rr,ss)) for (k,(rr,ss)) in RSE if 0<RIF(ss).endpoints()[1] and RIF(ss).endpoints()[0]<RIF(rr).endpoints()[1] and RIF(rr).endpoints()[0]<1] # 5875

#####################
# passe 2 : équations angles pc et mc :
RS2=passe2(RS1) # 601 en 9.43s
RS2E=passe2(RS1E) # 6

#####################
# passe 3 : existence grande couronne
RS3=passe3(RS2) # 175 en 3min 39s
RS3E=passe3(RS2E) # 1 (1) : 1rr1s/11rrs (28,76)
# erreur résistante (AssertionError) : on traite par variété
v=ideal(K,[eqpc[28],eqmc[76]]).variety(AA) # 45min, 29 couples, filtré à 1
c=RS3E[0]
RS3.append((c[0],filtre_racines(v)[0],c[2]))
RS3.sort()

#####################
# passe 4 : vérification exacte
E=passe4(RS)

# tout en 10min 58s sauf :
[(17, 135), (30, 50), (37, 12), (37, 119)]

# si on vire pc 10,13,17,37 : tous ok en 3min 36s
# 30,50=1r1r/11r1s : ok en substituant les valeurs exactes de r et s

