KX=PolynomialRing(QQ,['X1','X2','X3','X4','X5','X6','X7','X8','r','s'])
KX.inject_variables()
K=PolynomialRing(QQ,['r','s'])

def kashi(x,y,z):
    return ((y+x)^2+(y+z)^2-(x+z)^2)/2/(y+x)/(y+z)

# Al Kashi pour exprimer les cosinus comme fractions rationnelles en r et s
cosinus=[
kashi(1,s,1),kashi(1,s,r),kashi(1,s,s),kashi(r,s,r),kashi(r,s,s),kashi(s,s,s),
kashi(1,r,1),kashi(1,r,r),kashi(1,r,s),kashi(r,r,r),kashi(r,r,s),kashi(s,r,s),
kashi(1,1,1),kashi(1,1,r),kashi(1,1,s),kashi(r,1,r),kashi(r,1,s),kashi(s,1,s)
]

# les sinus via sqrt(1-cos^2) en utilisant les Xi pour la racine
sinus=[
2*X1/(s+1)^2, 2*X2/(s+1)/(r+s), X3/(s+1), 2*r*X4/(r+s)^2, X5/(r+s), X6/2,
2*X7/(r+1)^2, X8/(r+1), 2*X2/(r+1)/(r+s), X6/2, X4/(r+s), 2*s*X5/(r+s)^2,
X6/2, X7/(r+1), X1/(s+1), 2*r*X8/(r+1)^2, 2*X2/(r+1)/(s+1), 2*s*X3/(s+1)^2
]

# un facteur en r et s est inutile s'il n'a pas de racines dans le domaine 0<s<r<1
# ici heuristique : tous les coeffs du même signe ou quelques cas courants
def inutile(f):
    return len(set(sign(i) for i in f.coefficients()))==1 or f in [s-1,1-s,r-1,1-r,r-s,s-r]

# élimine les Xi de l'équation (long à cause étape élevation au carré)
dico=dict({X1:s*(s+2),X2:r*s*(r+s+1),X3:2*s+1,X4:s*(2*r+s),X5:r*(r+2*s),X6:3,X7:r*(r+2),X8:2*r+1})
def elimine(eqX):
    try:
        # vire les facteurs inutiles et, surtout, les multiplicités
        l=[f[0] for f in factor(eqX) if not set(f[0].variables()).issubset({r,s}) or not inutile(f[0])]
        if len(l)>1: # factorisation possible (tjrs facteur sans Xi ?)
            return prod([elimine(f) for f in l])
        eqX=prod(l)
        # choisit un Xi et l'élimine
        X=set(eqX.variables()).difference({r,s})
        Xi=X.pop()
        A=eqX.coefficient(Xi)
        B=eqX-A*Xi
        eqX=A^2*dico.get(Xi)-B^2 # c'est là que c'est long !
        # élimine les éventuels Xj^2 créés : ne reste que des Xj autre que Xi
        for Xj in X:
            eqX+=eqX.coefficient(Xj^2)*(dico.get(Xj)-Xj^2)
        return elimine(eqX)
    except KeyError: # plus de Xi à éliminer
        return eqX

# équation associé aux angles k, tc=0,1,2 pour pc, mc, gc ; variables Xi
def equationX(k,tc):
    (a,b,c,d,e,f)=SR.var('a b c d e f')
    v=(a,b,c,d,e,f)
    l=[cos(i) for i in v]+[sin(i) for i in v]
    eqX=KX(-1) # cos(2pi) du membre droit de l'équation
    if len([1 for i in k if i<>0])==1: # un seul angle : cas à part car operands(cos(a)) donne a et non cos(a)
        ops=[cos(v[i]) for i in range(6) if k[i]<>0]
    else: # cas normal
        ops=cos(sum(v[i]*sign(k[i]) for i in range(6))).trig_expand().operands()
    for o in ops:
        t=KX(1)
        for ft in o.factor_list():
            if ft[0]==-1:
                t=-t
            else:
                i=l.index(ft[0])
                if i<6: #cos
                    t*=chebyshev_T(k[i],cosinus[6*tc+i])
                else: #sin
                    t*=chebyshev_U(k[i-6]-1,cosinus[6*tc+i-6])*sinus[6*tc+i-6]
        eqX+=t
    return KX(numerator(eqX))

# équation sans les Xi
# marche avec k1-k2 !
def equation(k,tc):
    return K(elimine(equationX(k,tc)))

# vire les racines hors domaine (lorsque calculé via variété idéal)
def filtre_racines(v):
    z=[]
    for sol in v:
        s1=sol.items()[0][1]
        r1=sol.items()[1][1]
        if 0<s1 and s1<r1 and r1<1:
            z.append((r1,s1))
    return z

#################
## vérification ##

def elimine2(eqX,rs):
    try:
        # vire les facteurs inutiles et, surtout, les multiplicités
        l=[f[0] for f in factor(eqX) if not set(f[0].variables()).issubset({r,s}) or not inutile(f[0])]
        if len(l)>1: # factorisation possible (tjrs facteur sans Xi ?)
            return sum([elimine2(f,rs) for f in l])>0
        eqX=prod(l)
        # choisit un Xi et l'élimine
        X=set(eqX.variables()).difference({r,s})
        Xi=X.pop()
        A=eqX.coefficient(Xi)
        B=eqX-A*Xi
        # multiplication par AXi-B
        # vérif d'abord que AXi-B non nul en arithmétique d'intervalle
        C=A*Xi-B
        C=C.substitute({Xi:sqrt(SR(dico.get(Xi))) for Xi in set(eqX.variables()).difference({r,s})})
        C=RIF(C.substitute({r:RIF(rs[0]),s:RIF(rs[1])}))
        if C.contains_zero():
            raise Warning
        # multiplie par AXi-B (long)
        eqX=A^2*dico.get(Xi)-B^2
        # élimine les éventuels Xj^2 créés : ne reste que des Xj autre que Xi
        for Xj in X:
            eqX+=eqX.coefficient(Xj^2)*(dico.get(Xj)-Xj^2)
        return elimine2(eqX,rs)
    except KeyError: # plus de Xi à éliminer : vérifie exactement l'équation
        a=eqX.substitute({r:rs[0],s:rs[1]})
        a.exactify()
        return a==0

# c=((i,j),(r,s),(p,m,g))
# vérifie toutes les équations de couronne
# ie : équation en r et s annulée, mais n'annule aucun des facteurs AXi-B introduits pour éliminer les Xi
# si facteur AXi-B annulé : vérif directe sur eqX (sans éliminer les Xi), plus long
def verif_equation(c):
    for tc in [0,1,2]:
        for k in c[2][tc]:        
            eqX=equationX(k,tc)
            try:
                if not elimine2(eqX,c[1]):                    
                    return false
                print('+')
            except Warning: # parasite potentiel : substitue direct. ds les Xi sans les éliminer
                eq=eqX.substitute({Xi:sqrt(SR(dico.get(Xi))) for Xi in set(eqX.variables()).difference({r,s})})
                z=AA(eq.substitute(r=c[1][0],s=c[1][1]))
                z.exactify()
                if z<>0:
                    return false                
                print('parasite potentiel éliminé')
    print('(%d,%d) ok'%(c[0][0],c[0][1]))
    return true

##############################
## un fois qu'on a les solutions pc/mc, retrouver les valeurs de r et s
## pc et mc vecteurs d'angles
def pcmc2rs(pc,mc):
    eqpc=equation(pc,0)
    eqmc=equation(mc,1)
    R=ZZ[r](eqpc.resultant(eqmc,s))
    S=ZZ[s](eqpc.resultant(eqmc,r))
    RR=[]
    SS=[]
    for R in [f[0] for f in factor(R) if not inutile(f[0])]:
        RR+=R.roots(AA,multiplicities=false)
    for S in [f[0] for f in factor(S) if not inutile(f[0])]:
        SS+=S.roots(AA,multiplicities=false)
    RR=[rr for rr in RR if RIF(rr).endpoints()[0]<1]
    SS=[ss for ss in SS if 0<RIF(ss).endpoints()[1]]
    RS=flatten([[(rr,ss) for rr in RR if RIF(ss).endpoints()[0]<RIF(rr).endpoints()[1]] for ss in SS],max_level=1)
    RS=[(rr,ss) for (rr,ss) in RS if verif_couronne(pc,0,rr,ss) and verif_couronne(mc,1,rr,ss)]
    RS=[(rr,ss) for (rr,ss) in RS if len(couronnes_compatibles(rr,ss,2))>0]
    if len(RS)<>1:
        raise ValueError
    return RS[0]

